/*  @author : Michel Kinsy
 
 *  Copyright (c) 2012 Heracles (CSG/CSAIL/MIT)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <vector>
#include <map>
using namespace std;

typedef int ADDRESS; 
typedef string DATA;

typedef struct{
	string name; 
	ADDRESS hi_addr; 
	ADDRESS lo_addr; 
	int range; 
	vector<string> data;
} VARIABLE; 

typedef struct{
	string name;
	vector<string> lines; 
	vector<string> children; 
	vector<int> cores; 
} FUNCTION;

typedef struct{
	int core; 
	int cur_addr; 
	vector<string> vars; 
	vector<string> functs; 
	vector<string> text;
} WORKLOAD;

// User modifiable variables
char *filename;             // input file varaible
int default_core = 0; 
string mem_bits  = "16"; 

// Global variables
string token, line;
string target; 
string tokens;
int memory_pointer, 
counter = 0;

map<string, int> mem_set;
map<string, int> MIPS_regs;
map<ADDRESS, DATA> data_memory;
map<string, int> not_supported_ins;
map<string, int> ::iterator iter = MIPS_regs.begin();

vector<VARIABLE> vars; 
vector<FUNCTION> functions; 
vector<WORKLOAD> workloads;
vector<string> program_files;
vector<string> program_cores;
vector<int> active_cores; 

void scanner(void);
void register_set_up(void);
void instruction_correction(void); 

void trim(vector<int>& list);
void trim(vector<FUNCTION>& functions);
void workload_distribution(void);
void generate_workload(void);

void data_mapping(vector<string>& lines);
void generate_final_file (vector<string>& lines, int core); 

void phase1_transform(vector<string>& lines);
void phase2_transform(vector<string>& lines);

int main (int argc, char * const argv[]) {
	cout <<"--------------- MIPS-gcc processing ------------------\n";
	if (argc < 2) return 1; 
	if(argc >= 2) filename = argv[1];
	if(argc >= 3) mem_bits = argv[2];
	
	ostringstream argsout;
	if(argc == 4) argsout << argv[3];
	istringstream argsin (argsout.str());
	if(argc > 2) 	argsin >> default_core; 
 
	instruction_correction();
	register_set_up(); 
	scanner();
	generate_workload(); 
    	cout <<"------------- MIPS-gcc processing done! -------------\n";
	return 0;
}

void register_set_up(void)
{
	MIPS_regs["$zero"]=0; MIPS_regs["$0"]=0;    // constant 0
	MIPS_regs["$at"]= 1;  MIPS_regs["$1"]=1;    // reserved for assembler
	MIPS_regs["$v0"]= 2;  MIPS_regs["$v1"]=3;   // expression evaluation and function results
	MIPS_regs["$2"]= 2;   MIPS_regs["$3"]=3;
	MIPS_regs["$a0"]=4;   MIPS_regs["$a1"]=5;   MIPS_regs["$a2"]=6;
	MIPS_regs["$4"]=4;    MIPS_regs["$5"]=5;    MIPS_regs["$6"]=6;
	MIPS_regs["$a3"]=7;   MIPS_regs["$7"]=7;                      // arguments
	MIPS_regs["$t0"]=8;   MIPS_regs["$t1"]=9;   MIPS_regs["$t2"]=10;
	MIPS_regs["$8"]=8;    MIPS_regs["$9"]=9;    MIPS_regs["$10"]=10;
	MIPS_regs["$t3"]=11;  MIPS_regs["$t4"]=12;  MIPS_regs["$t5"]=13;
	MIPS_regs["$11"]=11;  MIPS_regs["$12"]=12;  MIPS_regs["$13"]=13;
	MIPS_regs["$t6"]=14;  MIPS_regs["$t7"]=15;   // temporary, saved by caller
	MIPS_regs["$14"]=14;  MIPS_regs["$15"]=15;
	MIPS_regs["$s0"]=16;  MIPS_regs["$s1"]=17;  MIPS_regs["$s2"]=18;
	MIPS_regs["$16"]=16;  MIPS_regs["$17"]=17;  MIPS_regs["$18"]=18;
	MIPS_regs["$s3"]=19;  MIPS_regs["$s4"]=20;  MIPS_regs["$s5"]=21;
	MIPS_regs["$19"]=19;  MIPS_regs["$20"]=20;  MIPS_regs["$21"]=21;
	MIPS_regs["$s6"]=22;  MIPS_regs["$s7"]=23;     // temporary, saved by called function
	MIPS_regs["$22"]=22;  MIPS_regs["$23"]=23;
	MIPS_regs["$t8"]= 24; MIPS_regs["$t9"]= 25 ;   // temporary, saved by caller
	MIPS_regs["$24"]= 24; MIPS_regs["$25"]= 25 ;
	MIPS_regs["$k0"]= 26; MIPS_regs["$k1"]= 27 ;   // reserved for kernel (OS)
	MIPS_regs["$26"]= 26; MIPS_regs["$27"]= 27 ;
	MIPS_regs["$gp"]= 28; MIPS_regs["$28"]= 28;    // points to middle of a 64K block in the data segment
	MIPS_regs["$sp"]= 29; MIPS_regs["$29"]= 29;    // stack pointer (top of stack)
	MIPS_regs["$fp"]= 30; MIPS_regs["$30"]= 30;    // frame pointer (beginning of current frame)
	MIPS_regs["$ra"]= 31; MIPS_regs["$31"]= 31;    // return address
}

void instruction_correction(void)
{
	not_supported_ins["li"]= 1;  not_supported_ins["la"]= 2;  not_supported_ins["b"]= 3; 
	not_supported_ins["move"]= 4; not_supported_ins["addu"]= 5;  not_supported_ins["subu"]= 6;
	not_supported_ins["slt"]= 7; not_supported_ins["sltu"]= 8; not_supported_ins["j"]= 9; 
	not_supported_ins["mult"]= 10; not_supported_ins["lb"]= 11; not_supported_ins["sb"]= 12;
	not_supported_ins["lui"]= 13;
	
	// Not supported at this time ...
	not_supported_ins["mul"]= 15;   not_supported_ins["mult"]= 15; 
	not_supported_ins["multu"]= 15; not_supported_ins["div"]= 15;
	not_supported_ins["divu"]= 15;  not_supported_ins["mfhi"]= 15; 
	not_supported_ins["mflo"]= 15;  not_supported_ins["mthi"]= 15; 
	not_supported_ins["mtlo"]= 15;  not_supported_ins["beqz"]= 15;
	// Not supported at all...
	not_supported_ins["bltzal"]= 15; not_supported_ins["bgezal"]= 15;
}

void trim(vector<FUNCTION>& functions){
	for (int i=0;i<functions.size(); i++) {
		std::sort(functions[i].children.begin(), functions[i].children.end());
		functions[i].children.erase(std::unique(functions[i].children.begin(),
		           functions[i].children.end()), functions[i].children.end());
	}	
}

void trim(vector<int>& list){
	std::sort(list.begin(), list.end());
	list.erase(std::unique(list.begin(),list.end()), list.end());
}

void update_functions(vector<FUNCTION>& functions, string name, int core){
	int index = -1; 
	for (int u = 0; u < functions.size(); u++){
		if(functions[u].name == name) {index = u ; break;}
	}
	
	if(index != -1) { 
		functions[index].cores.push_back(core);
		trim(functions[index].cores);
		for(int v = 0; v < functions[index].children.size(); v++) {
			string child = functions[index].children[v];
			if(child != functions[index].name) 
				update_functions(functions, child, core); 
		}
	} 

}

/* Extract variable placement and function-to-core mapping information */
void list_active_cores(void){
	string str; 
	int cur; 
	workloads.clear();  
	
	for (int i=0;i< functions.size(); i++) {
		str =  functions[i].name; 
		cur = 0; 
		while(str[cur] != '\0') {
			if(str[cur]== '_') str[cur]=' ';
			cur++;
		}
		token = ""; 
		istringstream local_list(str);
		local_list >> token; 
		if(token == "core") {
			local_list >> token;
			int cr = atoi(token.c_str()); 
			active_cores.push_back(cr); 
			update_functions(functions, functions[i].name, cr); 
		}
	}
	
	for (int i=0;i< functions.size(); i++) {
		cur = default_core;
		if(functions[i].cores.size() == 0) 
			functions[i].cores.push_back(cur); 
		cur = functions[i].cores[0];
		
		for (int j=0; j< functions[i].children.size(); j++)
			for (int k=0; k < functions.size(); k++)
			     if(functions[k].name == functions[i].children[j]) {
					functions[k].cores.push_back(cur);
					trim(functions[k].cores);
			     }
		
		trim(functions[i].cores); 
	}
}

void data_segment_allocations(void){
	for (int i=0;i< active_cores.size(); i++){
		WORKLOAD prog; 
		prog.core = active_cores[i]; 
		prog.cur_addr = 0;
		prog.text.clear(); 
		prog.cur_addr = 64; 
		prog.text.push_back(".space  32"); 
		
		for (int j=0; j< vars.size(); j++) {
			if(vars[j].hi_addr == active_cores[i]) {
				vars[j].lo_addr = prog.cur_addr;
				prog.cur_addr += vars[j].range;
				
				if(vars[j].data.size() == 0){
					stringstream ss; ss << vars[j].range;
					prog.text.push_back(".space " + ss.str()); 
				}
				else {
					for (int u=0; u< vars[j].data.size(); u++)
						prog.text.push_back(vars[j].data[u]); 
				}
				
				prog.vars.push_back(vars[j].name); 
			}
		}
		prog.text.push_back(".space  32");
		
		for (int j=0; j< functions.size(); j++) 
			for (int k=0; k< functions[j].cores.size(); k++) 
				if(functions[j].cores[k] == active_cores[i]) 
					prog.functs.push_back(functions[j].name);  
		
		workloads.push_back(prog); 
	}
}

void data_mapping2(void){
	vector<string> print_lines; 
	print_lines.push_back("\t.ent	h_print");
	print_lines.push_back("\t.type	h_print, @function");
	print_lines.push_back("h_print:");
	print_lines.push_back("\t.frame	$fp,8,$31		# vars= 0, regs= 1/0, args= 0, gp= 0");
	print_lines.push_back("\t.mask	0x40000000,-4");
	print_lines.push_back("\t.fmask	0x00000000,0");
	print_lines.push_back("\t.set	noreorder");
	print_lines.push_back("\t.set	nomacro");
	print_lines.push_back("\taddiu	$sp,$sp,-8");
	print_lines.push_back("\taddiu	$17,$4,0");
	print_lines.push_back("\taddiu	$2,$4,0");
	print_lines.push_back("\taddiu	$sp,$sp,8");
	print_lines.push_back("\tj	$31");
	print_lines.push_back("\tnop");
	print_lines.push_back("\t.set	macro");
	print_lines.push_back("\t.set	reorder");
	print_lines.push_back("\t.end	h_print");
	
	int cur;
	for (int i=0;i< functions.size(); i++) {
		vector<string> new_lines; 
		for (int j=0; j< functions[i].lines.size(); j++){
			string n_line = functions[i].lines[j]; 
			int adjust = 0; 
			cur = 0; 
			while(n_line[cur] != '\0') {
				if(n_line[cur]== '%') {
					string temp = n_line; 
					int index = 0; 
					while(temp[index] != '\0') {
						if((temp[index]== ',') |(temp[index]== ')') |(temp[index]== '('))
							temp[index] = ' ';
						index++; 
					}
					
					token = ""; 
					istringstream local_list(temp);
					int ind = 0; 
					int hi  = 0; 
					int add = 0; 
					string shift = "\tsll\t"; 
					string replace_line; 
					while (local_list >> token){
						if(ind == 0) {
                            replace_line = "\t"+token+"\t"; 
                            if (token == "addiu") add = 1; 
                        }
						if(ind == 1) {
							replace_line += token+","; 
							shift += token+","+mem_bits; 
						}
						if(ind == 2) {
							if(token =="%hi") hi = 1; 
							if(add == 1) replace_line += token+",";
						}
						if(ind == 3){
                            if(token =="%hi") hi = 1; 
                            if(add == 0){
    							for (int v=0; v< vars.size(); v++) 
    								if(vars[v].name == token) {
    									if(hi == 1) {
    										stringstream ss; ss << vars[v].hi_addr;
    										replace_line += ss.str();
    									}
    									else {
    										stringstream ss; ss << vars[v].lo_addr;
    										replace_line += ss.str();
    									}
    								}
							}
						}
						if(ind == 4) {
                            if(add == 1){
    							for (int v=0; v< vars.size(); v++) 
    								if(vars[v].name == token) {
    									if(hi == 1) {
    										stringstream ss; ss << vars[v].hi_addr;
    										replace_line += ss.str();
    									}
    									else {
    										stringstream ss; ss << vars[v].lo_addr;
    										replace_line += ss.str();
    									}
    								}
							}
							else replace_line += "("+token+")"; 
						}
						ind++; 
					}
					new_lines.push_back(replace_line); 
					if(hi == 1) new_lines.push_back(shift); 
					adjust = 1; 
				}
				cur++;
			}
			if(adjust == 0) new_lines.push_back(n_line); 
		}
		
		functions[i].lines.clear(); 
		if(functions[i].name == "h_print"){
			for (int j=0; j< print_lines.size(); j++)
				functions[i].lines.push_back(print_lines[j]);
		}
		else{
			for (int j=0; j< new_lines.size(); j++)
				functions[i].lines.push_back(new_lines[j]);
		}
	}
	
}

void workload_distribution(void){
	list_active_cores(); 
	trim(active_cores); 
	data_segment_allocations();  
	data_mapping2(); 
	
	for (int i=0;i< workloads.size(); i++){
		for (int u=0; u< workloads[i].functs.size(); u++) 
			for (int v=0; v< functions.size(); v++) 
				if(workloads[i].functs[u] == functions[v].name) {
					for (int j=0; j< functions[v].lines.size(); j++) 
						workloads[i].text.push_back(functions[v].lines[j]);
				}
	}
	
	program_files.clear(); program_cores.clear(); 
	cout<<" Program will be generated for cores: "; 
	for (int i=0;i< workloads.size(); i++)
		cout<< workloads[i].core << "  ";
	cout<<endl; 
}

int get_core(string var, vector<string>& lines){
	int ans = default_core; string tok; 
	for(int i = 0; i<lines.size(); i++){
		istringstream token_list(lines[i]);
		token_list >> tok;
		if(tok[tok.length()-1] == ':'){
			int u = 0; 
			 while(tok[u] != '\0') {
				if(tok[u]== '_') tok[u]=' ';
				u++;
			}
			istringstream t_list(tok); 
			string str; t_list>> str; 
			if(str =="core"){
				t_list>> str;
				ans = atoi(str.c_str()); 
			}
		}
		int index = 0; 
		if ((index = lines[i].find(var, index)) != string::npos)
			break; 
	}
	
	return ans; 
}

/* Scanning process and error cheching */
void scanner(void)
{
	vector<string> lines;
	vector<string> f_lines; 
	vector<string> f_children; 
	int valid_funct = 0; 
	int valid_var = 0;
	
	ifstream from;
	from.open(filename, ios::in);   /* open filename for reading */
	if(from.is_open())
		cin.rdbuf(from.rdbuf());  /* Redirect cin to the file reader buffer */
	else
	{
		cout << "Unable to open file.\n";  /* Explain the I/O error */
		exit(1);
	}

	/* Extracting the lines from the input file */
	active_cores.push_back(default_core); // by default core 0 is active. 
	while(getline(cin, line))
	{	
		istringstream token_list(line);
		token_list >> token;
		if (token ==".comm") {
			string str1, str2; 
			str1 = line; str2 = line;
			int u = 0; 
			
			while(str1[u] != '\0') {
				if(str1[u]== ',') str1[u]=' ';
				if(str2[u]== '_') str2[u]=' ';
				u++;
			} 
			
			istringstream list1(str1);
			istringstream list2(str2);
			
			VARIABLE var; 
			var.hi_addr = default_core; 
			var.lo_addr = 0; var.range = 0; 
			token = ""; 
			list1 >> token; list1 >> token;
			var.name = token; 
			if(list1 >> token) var.range = atoi(token.c_str()); 
			
			token = ""; 
			list2 >> token; list2 >> token;
			if(token == "core") {
				list2 >> token;
				int cr = atoi(token.c_str()); 
				active_cores.push_back(cr); 
				var.hi_addr = cr;
			}
			vars.push_back(var);
		}
		
		if (token ==".size") {
			token_list >> token;
			VARIABLE var; 
			var.hi_addr = default_core; 
			var.lo_addr = 0; var.range = 0; 
			var.name = token.substr(0, token.length()-1);
			valid_var = 0; 
			
			token = " "; 
			token_list >> token;
			if(token[0] != '.') {
				var.hi_addr = get_core(var.name, lines); 
				var.lo_addr = 0;
				var.range = atoi(token.c_str());
				vars.push_back(var);
				valid_var = 1;
			}
			
		}
		
		if (token ==".globl") {
			token_list >> token;
			FUNCTION func; 
			func.name = token; 
			functions.push_back(func);
		}
		
		if (token ==".ent") {
			f_lines.clear();
            f_children.clear();   
			valid_funct = 1; 
		}
		
		if (token =="jal") {
			token_list >> token; 
			if(token == "memcpy") line = ""; 
			else f_children.push_back(token);
		}
		
		if (token ==".end") {
			f_lines.push_back(line);
			valid_funct = 0; 
			if(functions.size()==0) {
				cout<<"ERROR parsing functions!!!\n";
				exit(1); 
			}
			else {
				int last = functions.size()-1; 
				for (int i=0;i<f_lines.size(); i++) {
					functions[last].lines.push_back(f_lines[i]);
				}
				for (int i=0;i<f_children.size(); i++) {
					functions[last].children.push_back(f_children[i]);
				}
			}
		}
		
		if(valid_funct) f_lines.push_back(line);
		if(valid_var == 3) {
			if((token != ".word") && (token != ".byte")) valid_var = 0; 
			else vars[vars.size()-1].data.push_back(line);
		}
		if(valid_var == 2) valid_var = 3;
		if(valid_var == 1) valid_var = 2; 
		
		lines.push_back(line);		
	}
	from.close();
	trim(functions);
	workload_distribution(); 
}

/* Reformat program files to the right ISA support */ 
void generate_workload(void){
	cout<<" Generating assembly file for cores ... \n"; 
	for (int i=0;i< workloads.size(); i++){
		phase1_transform(workloads[i].text);
		generate_final_file (workloads[i].text, workloads[i].core);
    }	
	
	remove("compile_log.txt");
	ofstream pos ("compile_log.txt", ios::out);
	pos<<"====== Compile Log: List of files to process  ====== \n";
    	for (int i=0; i< program_files.size(); i++) 
		pos<<program_files[i]<<" "<<program_cores[i]<<endl; 
	pos.close();
	
}

/* To remove the gcc MIPS environment set up information */
void phase1_transform(vector<string>& lines)
{ 
	int index= 0;
	vector<string> new_lines;
	
	/* Add buffering instructions to avoid executable instructions starting at address zero*/
	/* Assembly directives to avoid branch delay slot */
	new_lines.push_back(".space 32");
	new_lines.push_back(".set noreorder");
	new_lines.push_back(".set nomacro");
	new_lines.push_back(".set nomips16");
	new_lines.push_back(".set nomicromips");
	
	while(index < lines.size())
	{
		line = lines[index];
		istringstream token_list(line);
		token_list >> token;
		if (((token[0] =='.') && ((token !=".word") && (token !=".byte") && (token !=".space"))) || 
			(token =="gcc2_compiled.:") || (token =="__gnu_compiled_c:") || (token =="mflo") || 
			(token =="mfhi")) ; 
		else new_lines.push_back(line);
		index++;
	}
	
	new_lines.push_back("\tnop");
	
	for (int i=(new_lines.size()-2); i>0; i--) {
		istringstream token_list(new_lines[i]);
		token_list >>token; 
		if(token =="j" || token =="jr") {
			token_list >>token;
			if(token == "$31" || token == "$ra"){
				new_lines[i+1] = new_lines[i]; 
				new_lines[i] = "\taddi 	$17, $2, 0"; 
				break; 
			}
		} 
		else {
			new_lines[i+1] = new_lines[i]; 
		}
	}
	
	phase2_transform(new_lines);	
	lines.clear(); 
	for (int i=0; i< new_lines.size(); i++) 
		lines.push_back(new_lines[i]); 
}

/* To reformat the MIPS instructions to reflect the ISA version */
void phase2_transform(vector<string>& lines)
{
	vector<string> new_lines;
	int index= 0; string temp;
	int ins = 0; int num=0;
	int cur; int com;
	
	while(index < lines.size())
	{
		line = lines[index];
		istringstream token_list(line);
		if(token_list >> token) ins =  not_supported_ins[token];
		else ins = 0; 
		switch (ins) {
			case 1:
				temp= "\taddi ";
				cur= 2;
				while(line[cur] != ',') cur++;
				temp.append(line.substr(token.length()+1, cur-2));
				temp.append("$0");
				temp.append(line.substr(cur, line.length()));
				line = temp;
				break ;
			case 2:
				// have not decide the best way the manipulate la 
				break ;
			case 4:
				temp= "\taddi "; temp.append(line.substr(token.length()+1, line.length()));
				temp.append(",0");
				line = temp;
				break ;
			case 6:
				num=0;
				cur= 3;
				com = 0;
				while(line[cur] != '\0') {
					if(line[cur]== '$') num++;
					if(line[cur]== ',') com=cur;
					cur++;
				}
				if(num==2) {
					temp= "\taddiu ";
					temp.append(line.substr(token.length()+1, com-4));
					temp.append("-");
					temp.append(line.substr(com+1, line.length()));
					line = temp;
				}
				break ;
			case 9:
				token_list >> target;	  
				iter = MIPS_regs.find(target);
				if( iter != MIPS_regs.end() ) {
					temp= "\tjr "; 
					temp.append(line.substr(token.length()+1, line.length()));
					line = temp;
				}
				break ;
			case 10:
				cur= 3;
				num = 0;
				while(line[cur] != '\0') {
					if(line[cur]== '$') num=cur;
					if(line[cur]== ',') { 
						temp =line.substr(0, cur+1); 
						temp.append(line.substr(num, cur));
						temp.append(line.substr(cur+3, line.length()));
					}
					cur++;
				}
				line = temp;
				break ;
			case 13:
				temp= "\taddiu ";
				cur= 3;
				while(line[cur] != ',') cur++;
				temp.append(line.substr(token.length()+1, cur-3));
				temp.append("$0");
				temp.append(line.substr(cur, line.length()));
				line = temp;
				break ;
			case 15:
				cout<<"ERROR: Non-supported instuction!\n";  
				line = ".deadbeef";
				break ;
			default: 
				if((ins==3)||(ins==11)||(ins==12))
				{ 
					if (ins==3) 	temp= "\tj "; 
					if (ins==11) 	temp= "\tlw "; 
					if (ins==12) 	temp= "\tsw "; 
					temp.append(line.substr(token.length()+1, line.length()));
					line = temp;
				}
				if((ins==5)||(ins==7)||(ins==8))
				{  
					num=0;
					cur= 3;
					while(line[cur] != '\0') {if(line[cur]== '$') num++; cur++;}
					if(num==2) {
						if (ins==5) 	temp= "\taddiu ";
						if (ins==7) 	temp= "\tslti ";
						if (ins==8) 	temp= "\tsltiu ";
						temp.append(line.substr(token.length()+1, line.length()));
						line = temp;
					}
				};
		}
		new_lines.push_back(line);
		index++;
	}
	
	data_mapping(new_lines);	
	lines.clear(); 
	for (int i=0; i< new_lines.size(); i++) 
		lines.push_back(new_lines[i]); 
}

/* Data mapping section mapping to memory  */
void data_mapping (vector<string>& lines)
{
	int length = 0; string label, temp;
    for (int i=0; i< lines.size(); i++) 
	{
	    line = lines[i];
		istringstream tokens(line); 
		tokens >> label; length = label.length();
   		if (label[length-1]== ':') {
			label = label.substr(0,length-1);
			mem_set[label]= memory_pointer; counter= 0; 
   		}
		
		if ((label ==".word") ||(label ==".byte")) {
			counter++; tokens >> temp; 
			data_memory[memory_pointer]= temp; 
			memory_pointer += 4; 
		}
	}
}

int no_core_function(vector<string>& lines, string core_str){
	int ans = 1; int u; 
	for (int i=0; i< lines.size(); i++){
		u = 0; 
		if ((u = lines[i].find(core_str, u)) != string::npos){
			ans = 0; break; 
		}
	}
	return ans; 
}

void generate_final_file (vector<string>& lines, int core)
{
	stringstream s0, s1;
	string new_filename; 
	string core_str;
	string thread_str; 
	int u=0;
	s0 << filename; 
	s0 >> new_filename; 
	s1 << core;
	while (filename[u] !='.') u++;
	new_filename =  new_filename.substr(0, u);
	new_filename += "_c"+s1.str()+".asm"; 
	ofstream  pos (new_filename.c_str(), ios::out);
	
	pos<<"\n # Heracles supported MIPS Instructions \n";
    	for (int j=0; j< lines.size(); j++) 
		pos<<lines[j]<<endl;
	
	string dispath_str1 = "\taddiu 	$sp,$sp,-24\n\tsw	$31,20($sp)\n"; 
	dispath_str1 += "\tsw	$fp,16($sp)\n\taddi 	$fp,$sp,0"; 
	
	string stop_str = "Finish: \n\tnop\n\tnop\n\tnop\n\tnop\n\tj 	Finish\n\tnop\n";
	
	string dispath_str2 = "\tnop\n\tnop\n\taddi 	$sp,$fp,0\n\tlw	$31,20($sp)\n"; 
	dispath_str2 += "\tlw	$fp,16($sp)\n\taddiu 	$sp,$sp,24\n\tjr 	$31\n\tnop\n"; 
	
	core_str = "core_" + s1.str() +"_work"; 
	string dispath_core = "\n\tjal	"+ core_str + "\n";
	string dispath_main = "\n\tjal	main\n";
	
	thread_str = "core_" + s1.str() +"_work_T1";
	string dispath_thread = "\n\tjal	"+ thread_str + "\n";
	string stop_str1 = "Finish1: \n\tnop\n\tnop\n\tnop\n\tnop\n\tj 	Finish1\n\tnop\n";
	
	if(no_core_function(lines, core_str)) 
		pos<<core_str<<": \n"<<dispath_str1<<dispath_main<<stop_str<<dispath_str2<<endl;
		
	pos<<"\nDispatcher: \n"<<dispath_str1<<dispath_core<<stop_str<<dispath_str2<<endl; 
	
	if(no_core_function(lines, thread_str)== 0) 
		pos<<"\nDispatcher1: \n"<<dispath_str1<<dispath_thread<<stop_str1<<dispath_str2<<endl;
	else pos<<"\nDispatcher1: \n"<<dispath_str1<<"\n"<<stop_str1<<dispath_str2<<endl;
	
	pos.close();
	program_files.push_back(new_filename); 
	program_cores.push_back(s1.str());
}


